DECLARE SUB DrawScreen (V AS ANY, P AS ANY, XOffset%, YOffset%)
DECLARE SUB DoInput (P AS ANY)
DECLARE SUB LoadPalette (FileName$)
DECLARE SUB MSetPos (X%, Y%)
DECLARE SUB MRelLastPos (RX%, RY%)
DECLARE SUB MHide ()
DECLARE FUNCTION MInit% ()
DECLARE SUB Move (X%, Y%, Angle%, StepLen%)
DECLARE FUNCTION MULTIKEY% (T%)
DECLARE SUB InitTables ()
DECLARE SUB LoadMap (MapFile$, P AS ANY)
DECLARE SUB AlterViewPort (V AS ANY, XSize%, YSize%, FieldOfView%, RayStep%)
DECLARE SUB Parse (Text$, Separator$, WordCount%, Words$())
DEFINT A-Z
' $INCLUDE: 'QB.BI'
DIM SHARED Regs AS RegType

' -------------------------------------------------------------------
' RayCastR.BAS, a modular QuickBASIC raycaster v0.5
' by GenPFault (aka Charles Huber)
'
' email: genpfault@crosswinds.net
' webpage: http://crosswinds.net/~genpfault/
'
' Multikey routine by Joe Huber
' Parse SUB by Rob Davis
' LoadPalette SUB from PixelPlus256
' -------------------------------------------------------------------


' -------------------------------------------------------------------
'                             TYPE DEFS.
' -------------------------------------------------------------------

TYPE ViewPort  ' defines a view port for DrawScreen
 ScreenX AS INTEGER
 ScreenY AS INTEGER
 HalfScreenX AS INTEGER
 HalfScreenY AS INTEGER
 FOV AS INTEGER    ' (F)ield (O)f (V)ision
 HalfFOV AS INTEGER
 RayWidth AS INTEGER
 Dist2Plane AS INTEGER
 RayStep AS INTEGER   ' size of skip between columns
END TYPE

TYPE Player
 PlayerX AS INTEGER
 PlayerY AS INTEGER
 PlayerAng AS INTEGER
 PlayerAngInc AS INTEGER
 PlayerStep AS INTEGER    ' size of player step
 PlayerTilt AS INTEGER
END TYPE

' unused
TYPE KeyInput
 TurnLeft AS INTEGER
 TurnRight AS INTEGER
 Forward AS INTEGER
 Backward AS INTEGER
 StrafeLeft AS INTEGER
 StrafeRight AS INTEGER
 Strafe AS INTEGER
END TYPE

' unused
TYPE InputConfig
 Primary AS KeyInput
 Alt AS KeyInput
 MLook AS INTEGER
 MXFactor AS INTEGER   ' percent (0-100), neg #'s flip axis
 MYFactor AS INTEGER   ' ditto
END TYPE


' -------------------------------------------------------------------
'                         SHARED VARIABLES
' -------------------------------------------------------------------

DIM SHARED SinT2&(360), CosT2&(360) ' tables for look-ahead (eleminates a mul)
DIM SHARED SinT1&(360), CosT1&(360)
DIM SHARED SinT(360), CosT(360) ' SIN & COS tables

DIM SHARED ViewFix(-180 TO 180)   ' "fishbowl" remover table

DIM SHARED MaxDist
DIM SHARED MaxMapX, MaxMapY
'$DYNAMIC
DIM SHARED Map(0 TO 1, 0 TO 1)
'$STATIC


' -------------------------------------------------------------------
'                            CONSTANTS
' -------------------------------------------------------------------

CONST Multiplier = 8192
CONST GridSize = 64


' -------------------------------------------------------------------
'                          MAIN PROGRAM
' -------------------------------------------------------------------

SCREEN 13

LoadPalette "RAYCASTR.PAL"   ' loads gradient EGA palette
InitTables

' create & init view port
DIM VPort AS ViewPort
AlterViewPort VPort, 320, 200 - 8, 60, 2

' create & init player
DIM P AS Player
P.PlayerAngInc = 4
P.PlayerStep = 10

' load map
LoadMap "map1.txt", P

' init input devices
IF NOT MInit THEN END
MHide
MSetPos 0, 0
Z = MULTIKEY(-1)

DO

 DoInput P

 DrawScreen VPort, P, 0, 8

 F = F + 1
 IF TIMER - T# > 1 THEN
  FPS = F
  LOCATE 1, 1: PRINT "FPS:"; FPS
  F = 0
  T# = TIMER
 END IF

 ' uncomment for kewl rear-view mirror effect (WARNING, Giga-flicker!)
 'AlterViewPort VPort, 100, 50, 60, 4
 'P.PlayerAng = P.PlayerAng + 180
 'DrawScreen VPort, P, 110, 0
 'P.PlayerAng = P.PlayerAng - 180
 'AlterViewPort VPort, 320, 200, 60, 4

 ' wait for vertical retrace
 'WAIT &H3DA, 8
 'WAIT &H3DA, 8, 8

LOOP UNTIL MULTIKEY(1)

' unload keyboard handler
Z = MULTIKEY(-2)

END

SUB AlterViewPort (V AS ViewPort, XSize, YSize, FieldOfView, RayStep)

' stuffs a ViewPort type with given data

V.ScreenX = XSize
V.ScreenY = YSize
V.HalfScreenX = V.ScreenX \ 2
V.HalfScreenY = V.ScreenY \ 2
V.FOV = FieldOfView
V.HalfFOV = V.FOV \ 2
V.RayWidth = (V.FOV / V.ScreenX) * Multiplier
V.Dist2Plane = V.HalfScreenX / TAN(V.HalfFOV * 3.14159 / 180)
V.RayStep = RayStep

END SUB

SUB DoInput (P AS Player)

 ' this is kludgey as hell.  In a real game, this would be much better

 OldPlayerX = P.PlayerX
 OldPlayerY = P.PlayerY

 IF MULTIKEY(56) THEN Strafe = 1 ELSE Strafe = 0

 MRelLastPos MRelX, MRelY
 IF Strafe THEN
  Move P.PlayerX, P.PlayerY, P.PlayerAng + 90, -(MRelX * 2)
 ELSE
  P.PlayerAng = P.PlayerAng + (MRelX \ 2)
 END IF
 'P.PlayerTilt = P.PlayerTilt - MRelY
 Move P.PlayerX, P.PlayerY, P.PlayerAng, (MRelY * 2)
 IF P.PlayerX < 0 THEN P.PlayerX = 0
 IF P.PlayerY < 0 THEN P.PlayerY = 0
 IF P.PlayerX > MaxMapX THEN P.PlayerX = MaxMapX
 IF P.PlayerY > MaxMapY THEN P.PlayerY = MaxMapY

 IF MULTIKEY(80) THEN Move P.PlayerX, P.PlayerY, P.PlayerAng, P.PlayerStep
 IF MULTIKEY(72) THEN Move P.PlayerX, P.PlayerY, P.PlayerAng, -P.PlayerStep
 IF MULTIKEY(77) THEN
  IF Strafe THEN
   Move P.PlayerX, P.PlayerY, P.PlayerAng - 90, P.PlayerStep
  ELSE
   P.PlayerAng = P.PlayerAng + P.PlayerAngInc
  END IF
 END IF
 IF MULTIKEY(75) THEN
  IF Strafe THEN
   Move P.PlayerX, P.PlayerY, P.PlayerAng + 90, P.PlayerStep
  ELSE
   P.PlayerAng = P.PlayerAng - P.PlayerAngInc
  END IF
 END IF
 IF Map(P.PlayerX \ GridSize, P.PlayerY \ GridSize) THEN
  ' oops, hit a wall, slide
  IF Map(P.PlayerX \ GridSize, OldPlayerY \ GridSize) THEN
   P.PlayerX = OldPlayerX
  END IF
  IF Map(OldPlayerX \ GridSize, P.PlayerY \ GridSize) THEN
   P.PlayerY = OldPlayerY
  END IF
 END IF

 P.PlayerAng = (P.PlayerAng + 360) MOD 360

END SUB

SUB DrawScreen (V AS ViewPort, P AS Player, XOffset, YOffset)

' Renders a view based on a ViewPort & Player structures
' X & YOffset simply move the image around the screen

Angle = (P.PlayerAng - V.HalfFOV + 360) MOD 360
TestAngle = Angle
AngleStep = V.RayWidth * V.RayStep
VFix = V.HalfFOV
M& = Multiplier

' weird gradient floor thingie
'C = (15 * 16) + 15
'Diff = ABS((V.ScreenY - V.HalfScreenY) - (V.ScreenY + YOffset))
'Inc = Diff / 16
'Y = V.ScreenY - V.HalfScreenY + YOffset
'DO
' LINE (XOffset, Y)-(V.ScreenX + XOffset, Y + Inc), C, BF
' Y = Y + Inc
' C = C - 1
'LOOP UNTIL Y >= V.ScreenY + YOffset + Inc

FOR I = 0 TO V.ScreenX - V.RayStep STEP V.RayStep
 A = (TestAngle& \ Multiplier + Angle) MOD 360
 TestX& = P.PlayerX
 TestY& = P.PlayerY
 ' cast out high grainularity rays
 SinVal& = SinT2&(A): CosVal& = CosT2&(A)
 DO
  TestX& = TestX& - CosVal&: TestY& = TestY& - SinVal&
  TX = (TestX& \ Multiplier) + P.PlayerX
  TY = (TestY& \ Multiplier) + P.PlayerY
  MapColor = Map(TX \ GridSize, TY \ GridSize)
 LOOP UNTIL MapColor
 TestX& = TestX& + CosVal&: TestY& = TestY& + SinVal&
 ' cast out mid grainularity rays
 SinVal& = SinT1&(A): CosVal& = CosT1&(A)
 DO
  TestX& = TestX& - CosVal&: TestY& = TestY& - SinVal&
  TX = (TestX& \ Multiplier) + P.PlayerX
  TY = (TestY& \ Multiplier) + P.PlayerY
  MapColor = Map(TX \ GridSize, TY \ GridSize)
 LOOP UNTIL MapColor
 TestX& = TestX& + CosVal&: TestY& = TestY& + SinVal&
 ' cast out low grainularity rays
 SinVal& = SinT(A): CosVal& = CosT(A)
 DO
  TestX& = TestX& - CosVal&: TestY& = TestY& - SinVal&
  TX = (TestX& \ Multiplier) + P.PlayerX
  TY = (TestY& \ Multiplier) + P.PlayerY
  MapColor = Map(TX \ GridSize, TY \ GridSize)
 LOOP UNTIL MapColor

 ' Use SIN or COS for distance calculation, based on angle
 ' Orginally I just used COS, but as the angle approached 90 & 270,
 ' wierd things happened to the walls as COS became very small.
 IF (A > 45 AND A < 135) OR (A > 225 AND A < 315) THEN
  Dist& = ABS((P.PlayerY - TY) * M& \ SinT(A))
 ELSE
  Dist& = ABS((P.PlayerX - TX) * M& \ CosT(A))
 END IF
 Dist& = (Dist& * ViewFix(VFix& \ Multiplier + V.HalfFOV)) \ Multiplier + 1
 SliceHeight = ABS((GridSize / Dist& * V.Dist2Plane) \ 2)
 MapColor = MapColor + (Dist& / MaxDist * 16)

 ' Precalc line (box, really) endpoints
 XPos1 = I + XOffset
 XPos2 = XPos1 + V.RayStep - 1
 YPos1 = YOffset + V.HalfScreenY + P.PlayerTilt - SliceHeight
 YPos2 = YOffset + V.HalfScreenY + P.PlayerTilt + SliceHeight

 ' clip slice to view port
 IF YPos1 < YOffset THEN YPos1 = YOffset
 IF YPos2 > V.ScreenY + YOffset THEN YPos2 = V.ScreenY + YOffset

 ' Draw sky
 LINE (XPos1, YOffset)-(XPos2, YPos1), 0, BF
 ' Draw wall strip
 LINE (XPos1, YPos1)-(XPos2, YPos2), MapColor, BF
 ' Draw ground
 LINE (XPos1, YPos2)-(XPos2, YOffset + V.ScreenY - 1), 0, BF

 TestAngle& = TestAngle& + AngleStep
 VFix& = VFix& - AngleStep
NEXT I

END SUB

SUB InitTables

'Create SIN and COS tables.

Step2& = GridSize \ 2
Step1& = GridSize \ 4

FOR Angle = 0 TO 360
 SinT2&(Angle) = SIN(Angle * .017453) * Multiplier * Step2&
 CosT2&(Angle) = COS(Angle * .017453) * Multiplier * Step2&
 SinT1&(Angle) = SIN(Angle * .017453) * Multiplier * Step1&
 CosT1&(Angle) = COS(Angle * .017453) * Multiplier * Step1&
 SinT(Angle) = SIN(Angle * .017453) * Multiplier
 CosT(Angle) = COS(Angle * .017453) * Multiplier
NEXT Angle

FOR Angle = -180 TO 180
 ViewFix(Angle) = COS(Angle * .017453) * Multiplier
NEXT Angle

END SUB

SUB KEYTEST

SCREEN 0
CLS

Z = MULTIKEY(-1)
DO
X = 1
Y = 1

 FOR I = 1 TO 128
 
  TEST = MULTIKEY(I)
  LOCATE Y, X
  PRINT USING "## =###"; TEST; I

  IF Y < 23 THEN
   Y = Y + 1
  ELSE
   Y = 1
   X = X + 9
  END IF

 NEXT I

LOOP WHILE MULTIKEY(1) = 0

Z = MULTIKEY(-2)

END

END SUB

SUB LoadMap (MapFile$, P AS Player)

F = FREEFILE
OPEN MapFile$ FOR INPUT AS F

DIM Words$(100)

MapY = 0
DO WHILE NOT EOF(F)
 LINE INPUT #F, A$
 Parse A$, CHR$(32), NumWords, Words$()
 SELECT CASE UCASE$(Words$(1))
  CASE "MAPSIZE"
   XDim = VAL(Words$(2))
   YDim = VAL(Words$(3))
   REDIM Map(0 TO XDim, 0 TO YDim)
   MaxMapX = GridSize * XDim
   MaxMapY = GridSize * YDim
  CASE "ANGLE"
   P.PlayerAng = VAL(Words$(2))
  CASE "D"
   FOR X = 0 TO NumWords - 2
    C = VAL(Words$(X + 2))
    IF C = -1 THEN
     P.PlayerX = (X * GridSize) + (GridSize \ 2)
     P.PlayerY = (MapY * GridSize) + (GridSize \ 2)
     Map(X, MapY) = 0
    ELSE
     Map(X, MapY) = C * 16
    END IF
   NEXT X
   MapY = MapY + 1
 END SELECT
LOOP

CLOSE F

A& = MaxMapX
B& = MaxMapY
MaxDist = INT(SQR(A& * A& + B& * B&))

END SUB

SUB LoadPalette (FileName$)

    'Size array to hold all 256 colours.
    REDIM PaletteArray&(0 TO 255)

    IF FileName$ <> "" THEN
	'*** Read palette data from file ***
	FileNo = FREEFILE
	OPEN FileName$ FOR BINARY AS #FileNo
	FOR n = 0 TO 255
	    GET #FileNo, , colour&
	    PaletteArray&(n) = colour&
	NEXT n
	CLOSE #FileNo
    ELSE
	'*** Read palette data from DATA statements ***
	FOR n = 0 TO 255
	    READ colour&
	    PaletteArray&(n) = colour&
	NEXT n
    END IF
  
    'Break down all 256 colours into their RGB values.
    DIM RGBval(0 TO 255, 0 TO 2)
    FOR n = 0 TO 255
	C& = PaletteArray&(n)
	B = C& \ 65536: C& = C& - B * 65536
	g = C& \ 256: C& = C& - g * 256
	r = C&
	RGBval(n, 0) = r
	RGBval(n, 1) = g
	RGBval(n, 2) = B
    NEXT n

    'Write colours directly to the video card.
    WAIT &H3DA, &H8, &H8: WAIT &H3DA, &H8
    FOR n = 0 TO 255
	OUT &H3C8, n            'Select attribute.
	OUT &H3C9, RGBval(n, 0) 'Write red.
	OUT &H3C9, RGBval(n, 1) 'Write green.
	OUT &H3C9, RGBval(n, 2) 'Write blue.
    NEXT n

END SUB

SUB MHide
 
 ' Hides mouse

 Regs.ax% = 2
 INTERRUPT &H33, Regs, Regs

END SUB

FUNCTION MInit  ' Initializes mouse

 ' Returns -1 if mouse is present, 0 if it isn't

 Regs.ax% = 0
 INTERRUPT &H33, Regs, Regs

 MInit = Regs.ax%

END FUNCTION

SUB Move (X, Y, Angle, StepLen)

S& = StepLen
A = (Angle + 360) MOD 360
X = X + (CosT(A) * S&) \ Multiplier
Y = Y + (SinT(A) * S&) \ Multiplier

END SUB

SUB MRelLastPos (RX, RY)

 ' Returns location of mouse relative to last postion
 ' i.e., if both numbers are negitive, the mouse was moved
 ' left & up; positive, down & right
 ' (A use for SGN()! See Demo4 )
	       
 Regs.ax% = 11
 INTERRUPT &H33, Regs, Regs
 RX = Regs.cx%
 RY = Regs.dx%
  
END SUB

SUB MSetPos (X, Y)

 ' Sets mouse position
 ' X = horizontal mouse positon
 ' Y = vertical mouse position

 Regs.ax% = 4
 Regs.cx% = X
 Regs.dx% = Y
 INTERRUPT &H33, Regs, Regs
 
END SUB

SUB MShow

 ' Shows mouse

 Regs.ax% = 1
 INTERRUPT &H33, Regs, Regs

END SUB

SUB MStatus (X, Y, B1, B2)

 ' Returns location of mouse & status of buttons

 Regs.ax% = 3
 INTERRUPT &H33, Regs, Regs
 X = Regs.cx%
 Y = Regs.dx%
 B1 = (Regs.bx% AND 1) <> 0
 B2 = (Regs.bx% AND 2) <> 0

END SUB

FUNCTION MULTIKEY (T)

STATIC kbcontrol(), kbmatrix(), Firsttime, StatusFlag

IF Firsttime = 0 THEN          'Initalize
 DIM kbcontrol(128)
 DIM kbmatrix(128)
 code$ = ""
 code$ = code$ + "E91D00E93C00000000000000000000000000000000000000000000000000"
 code$ = code$ + "00001E31C08ED8BE24000E07BF1400FCA5A58CC38EC0BF2400B85600FAAB"
 code$ = code$ + "89D8ABFB1FCB1E31C08EC0BF2400BE14000E1FFCFAA5A5FB1FCBFB9C5053"
 code$ = code$ + "51521E560657E460B401A8807404B400247FD0E088C3B700B0002E031E12"
 code$ = code$ + "002E8E1E100086E08907E4610C82E661247FE661B020E6205F075E1F5A59"
 code$ = code$ + "5B589DCF"
 DEF SEG = VARSEG(kbcontrol(0))
 FOR I = 0 TO 155                     ' Load ASM
     d = VAL("&h" + MID$(code$, I * 2 + 1, 2))
     POKE VARPTR(kbcontrol(0)) + I, d
 NEXT I
 I& = 16       ' I think this stuff connects the interrupt with kbmatrix()
 n& = VARSEG(kbmatrix(0)): l& = n& AND 255: h& = ((n& AND &HFF00) \ 256): POKE I&, l&: POKE I& + 1, h&: I& = I& + 2
 n& = VARPTR(kbmatrix(0)): l& = n& AND 255: h& = ((n& AND &HFF00) \ 256): POKE I&, l&: POKE I& + 1, h&: I& = I& + 2
 DEF SEG
 Firsttime = 1
END IF

SELECT CASE T
 CASE -1
  IF StatusFlag = 0 THEN
   DEF SEG = VARSEG(kbcontrol(0))
   CALL ABSOLUTE(0)                     ' Run interrupt
   DEF SEG
   StatusFlag = 1
  END IF
 CASE -2
  IF StatusFlag = 1 THEN
   DEF SEG = VARSEG(kbcontrol(0))      ' Turn off interrupt
   CALL ABSOLUTE(3)
   DEF SEG
   StatusFlag = 0
  END IF
 CASE 1 TO 128
  MULTIKEY = kbmatrix(T)               ' Return status
 CASE ELSE
  MULTIKEY = 0                          ' User Supidity Error
END SELECT

END FUNCTION

SUB Parse (Text$, Separator$, WordCount%, Words$())

' procedure to return Text$ as individual words in the array Words$()
' where Separator$ is the character which identifies the separator
' between the words - this is normally a space, or CHR$(32)
' the number of words found is returned as WordCount%

' create space for up to 100 words - amend to suit
REDIM Words$(100)

' remove any leading or trailing spaces
Text$ = LTRIM$(RTRIM$(Text$))

' zero the counter of words
WordCount% = 0

' if there is a separator, start processing;
' else, return the entire contents of Text$
' as the first word of the array

IF INSTR(Text$, Separator$) > 0 THEN

	' continue to loop whilst there are some separators
	DO WHILE INSTR(Text$, Separator$) > 0
	
		' there is a separator so increment the wordcount
		WordCount% = WordCount% + 1
	
		' find out where the next separator occurs in Text$
		Place% = INSTR(Text$, Separator$)
	
		' strip the leftmost characters up to but not including
		' the separator
		Words$(WordCount%) = LEFT$(Text$, Place% - 1)
	
		' reduce Text$ to what's left after and not including the
		'separator
		Text$ = MID$(Text$, Place% + 1)
	LOOP
ELSE
	WordCount% = 1
	Words$(WordCount%) = Text$
END IF

END SUB

